# standard libraries
import copy
import datetime
import io
import json
import os
import pathlib
import typing
import uuid
import zipfile
import itertools

# third party libraries
import imageio.v3 as imageio
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.data import Image
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import Utility
from nion.utils import DateTime


DataElementType = typing.Dict[str, typing.Any]
_DataArrayType = numpy.typing.NDArray[typing.Any]


class ImportExportIncompatibleDataError(Exception):
    pass


class ImportExportHandler:

    """
        A base class for implementing import/export handlers.

        :param name: the localized name for the handler; will appear in file dialogs
        :param extensions: the list of handled extensions; do not include leading dots
    """

    # Extensions should not include a period.
    def __init__(self, io_handler_id: str, name: str, extensions: typing.Sequence[str]) -> None:
        self.io_handler_id = io_handler_id
        self.name = name
        self.extensions = list(extensions)
        self.supports_composite_data = False

    def can_read(self) -> bool:
        return True

    # return data items
    def read_data_items(self, extension: str, file_path: pathlib.Path) -> typing.Sequence[DataItem.DataItem]:
        data_items: typing.List[DataItem.DataItem] = list()
        if file_path.exists() or str(file_path.parent).startswith(":"):
            data_items.extend(self._read_data_items(extension, file_path))
        return data_items

    def _read_data_items(self, extension: str, file_path: pathlib.Path) -> typing.Sequence[DataItem.DataItem]:
        data_items = list()
        data_elements = self.read_data_elements(extension, file_path)
        for data_element in data_elements:
            if "data" in data_element:
                if not "title" in data_element:
                    title = file_path.stem
                    data_element["title"] = title
                data_element["filepath"] = file_path
                data_item = create_data_item_from_data_element(data_element, file_path)
                data_items.append(data_item)
        return data_items

    # return data
    def read_data_elements(self, extension: str, path: pathlib.Path) -> typing.Sequence[DataElementType]:
        return list()

    def can_write(self, data_metadata: DataAndMetadata.DataMetadata, extension: str) -> bool:
        return False

    def write_display_item(self, display_item: DisplayItem.DisplayItem, path: pathlib.Path, extension: str) -> None:
        data_item = display_item.data_item
        assert data_item
        with open(path, 'wb') as f:
            data = data_item.data
            if data is not None:
                self.write_data(data, extension, f)

    def write_data(self, data: _DataArrayType, extension: str, file: typing.BinaryIO) -> None:
        pass


class ImportExportManager(metaclass=Utility.Singleton):
    """
        Tracks import/export plugins.
    """
    def __init__(self) -> None:
        # we store a of dicts dicts containing extensions,
        # load_func, save_func, keyed by name.
        self.__io_handlers: typing.List[ImportExportHandler] = []

    def register_io_handler(self, io_handler: ImportExportHandler) -> None:
        self.__io_handlers.append(io_handler)

    def unregister_io_handler(self, io_handler: ImportExportHandler) -> None:
        self.__io_handlers.remove(io_handler)

    def get_readers(self) -> typing.Sequence[ImportExportHandler]:
        readers = []
        for io_handler in self.__io_handlers:
            if io_handler.can_read():
                readers.append(io_handler)
        return readers

    def get_writers(self) -> typing.Sequence[ImportExportHandler]:
        writers = []
        for io_handler in self.__io_handlers:
            if io_handler.can_read():
                writers.append(io_handler)
        return writers

    def get_writer_by_id(self, io_handler_id: str) -> typing.Optional[ImportExportHandler]:
        for io_handler in self.__io_handlers:
            if io_handler.io_handler_id == io_handler_id:
                return io_handler
        return None

    def get_writers_for_data_item(self, data_item: DataItem.DataItem) -> typing.Sequence[ImportExportHandler]:
        writers = []
        data_metadata = data_item.data_metadata
        if data_metadata:
            for io_handler in self.__io_handlers:
                for extension in io_handler.extensions:
                    if io_handler.can_write(data_metadata, extension.lower()):
                        writers.append(io_handler)
        return writers

    def get_writers_for_display_item(self, display_item: DisplayItem.DisplayItem) -> typing.Sequence[ImportExportHandler]:
        writers = self.get_writers_for_data_item(display_item.data_items[0]) if display_item.data_items else list()
        if len(display_item.data_items) > 1:
            composite_writers = list()
            for writer in writers:
                if writer.supports_composite_data:
                    composite_writers.append(writer)
            writers = composite_writers
        return writers

    # read file, return data items
    def read_data_items(self, path: pathlib.Path) -> typing.Sequence[DataItem.DataItem]:
        root, extension = os.path.splitext(path)
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            for io_handler in self.__io_handlers:
                if extension in io_handler.extensions:
                    return io_handler.read_data_items(extension, path)
        return list()

    # read file, return data elements
    def read_data_elements(self, path: pathlib.Path) -> typing.Sequence[DataElementType]:
        root, extension = os.path.splitext(path)
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            for io_handler in self.__io_handlers:
                if extension in io_handler.extensions:
                    return io_handler.read_data_elements(extension, path)
        return list()

    def write_display_item_with_writer(self, writer: ImportExportHandler, display_item: DisplayItem.DisplayItem, path: pathlib.Path) -> None:
        extension = path.suffix
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            data_metadata = display_item.data_items[0].data_metadata if display_item.data_items else None
            if extension in writer.extensions and data_metadata and writer.can_write(data_metadata, extension):
                writer.write_display_item(display_item, path, extension)

    def write_display_item(self, display_item: DisplayItem.DisplayItem, path: pathlib.Path) -> None:
        extension = path.suffix
        if extension:
            extension = extension[1:]  # remove the leading "."
            extension = extension.lower()
            for io_handler in self.__io_handlers:
                data_item = display_item.data_item
                assert data_item
                data_metadata = data_item.data_metadata
                if extension in io_handler.extensions and data_metadata and io_handler.can_write(data_metadata, extension):
                    io_handler.write_display_item(display_item, path, extension)


# create a new data item with a data element.
# data element is a dict which can be processed into a data item
# when this method returns, the data item has not been added to a document. therefore, the
# data is still loaded into memory, but with a data ref count of zero.
def create_data_item_from_data_element(data_element: DataElementType,
                                       data_file_path: typing.Optional[pathlib.Path] = None) -> DataItem.DataItem:
    uuid_str = data_element.get("uuid")
    uuid_ = uuid.UUID(uuid_str) if uuid_str else None
    large_format = data_element.get("large_format")
    if large_format is None:
        data = data_element.get("data")
        large_format = len(data.shape) > 2 and data.dtype != numpy.uint8 if data is not None else False
    data_item = DataItem.DataItem(item_uuid=uuid_, large_format=large_format)
    update_data_item_from_data_element(data_item, data_element, data_file_path)
    return data_item


# update an existing data item with a data element.
# data element is a dict which can be processed into a data item
# the existing data item may have a new size and dtype after returning.
def update_data_item_from_data_element(data_item: DataItem.DataItem, data_element: DataElementType,
                                       data_file_path: typing.Optional[pathlib.Path] = None) -> None:
    version = data_element["version"] if "version" in data_element else 1
    if version == 1:
        update_data_item_from_data_element_1(data_item, data_element, data_file_path)
    else:
        raise NotImplementedError("Data element version {:d} not supported.".format(version))


def update_data_item_from_data_element_1(data_item: DataItem.DataItem, data_element: DataElementType,
                                         data_file_path: typing.Optional[pathlib.Path] = None) -> None:
    assert data_item
    with data_item.data_item_changes(), data_item.data_source_changes():
        # file path
        # master data
        if data_file_path is not None:
            data_item.source_file_path = data_file_path
        data_and_metadata = convert_data_element_to_data_and_metadata(data_element)
        dimensional_calibrations = data_and_metadata.dimensional_calibrations
        intensity_calibration = data_and_metadata.intensity_calibration
        is_sequence = data_and_metadata.is_sequence
        collection_dimension_count = data_and_metadata.collection_dimension_count
        datum_dimension_count = data_and_metadata.datum_dimension_count
        data_shape_data_dtype = data_and_metadata.data_shape_and_dtype
        assert data_shape_data_dtype is not None
        is_same_shape = data_item.data_shape == data_shape_data_dtype[0] and data_item.data_dtype == data_shape_data_dtype[1] and data_item.is_sequence == is_sequence and data_item.collection_dimension_count == collection_dimension_count and data_item.datum_dimension_count == datum_dimension_count
        if is_same_shape:
            data_element_data = data_and_metadata.data
            assert data_element_data is not None
            with data_item.data_ref() as data_ref:
                sub_area = data_element.get("sub_area")
                data = data_ref.data
                if data is not None:
                    if sub_area is not None:
                        top = sub_area[0][0]
                        bottom = sub_area[0][0] + sub_area[1][0]
                        left = sub_area[0][1]
                        right = sub_area[0][1] + sub_area[1][1]
                        data[top:bottom, left:right] = data_element_data[top:bottom, left:right]
                    else:
                        data[:] = data_element_data[:]
                data_ref.data_updated()  # trigger change notifications
            if dimensional_calibrations is not None:
                for dimension, dimensional_calibration in enumerate(dimensional_calibrations):
                    data_item.set_dimensional_calibration(dimension, dimensional_calibration)
            if intensity_calibration:
                data_item.set_intensity_calibration(intensity_calibration)
            data_item.metadata = data_and_metadata.metadata
        else:
            data_item.set_xdata(data_and_metadata)
        # title
        if "title" in data_element:
            data_item.title = data_element["title"]
        # description
        # dates are _local_ time and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
        # time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
        # daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
        # timezone is for conversion and is the Olson timezone string.
        # datetime.datetime.strptime(datetime.datetime.isoformat(datetime.datetime.now()), "%Y-%m-%dT%H:%M:%S.%f" )
        # datetime_modified, datetime_modified_tz, datetime_modified_dst, datetime_modified_tzname is the time at which this image was modified.
        # datetime_original, datetime_original_tz, datetime_original_dst, datetime_original_tzname is the time at which this image was created.
        utc_datetime = data_and_metadata.timestamp
        data_item.created = utc_datetime
        if "time_zone" in data_and_metadata.metadata.get("description", dict()):
            timezone_dict = copy.deepcopy(data_and_metadata.metadata["description"]["time_zone"])
            timezone = timezone_dict.get("timezone")
            if timezone is not None:
                data_item.timezone = timezone
            timezone_offset = timezone_dict.get("tz")
            if timezone_offset is not None:
                data_item.timezone_offset = timezone_offset
        # author
        # sample
        # facility
        # location
        # gps
        # instrument
        # copyright
        # exposure
        # extra_high_tension


def convert_data_element_to_data_and_metadata(data_element: DataElementType) -> DataAndMetadata.DataAndMetadata:
    # NOTE: takes ownership of data_element['data']
    version = data_element["version"] if "version" in data_element else 1
    if version == 1:
        return convert_data_element_to_data_and_metadata_1(data_element)
    else:
        raise NotImplementedError("Data element version {:d} not supported.".format(version))


def convert_data_element_to_data_and_metadata_1(data_element: DataElementType) -> DataAndMetadata.DataAndMetadata:
    """Convert a data element to xdata. No data copying occurs.

    The data element can have the following keys:
        data (required)
        is_sequence, collection_dimension_count, datum_dimension_count (optional description of the data)
        spatial_calibrations (optional list of spatial calibration dicts, scale, offset, units)
        intensity_calibration (optional intensity calibration dict, scale, offset, units)
        metadata (optional)
        properties (get stored into metadata.hardware_source)
        one of either timestamp or datetime_modified
        if datetime_modified (dst, tz) it is converted and used as timestamp
            then timezone gets stored into metadata.description.timezone.
    """
    # data. takes ownership.
    data = data_element["data"]
    dimensional_shape = Image.dimensional_shape_from_data(data)
    is_sequence = data_element.get("is_sequence", False)
    dimension_count = len(dimensional_shape) if dimensional_shape else 0
    adjusted_dimension_count = dimension_count - (1 if is_sequence else 0)
    collection_dimension_count = data_element.get("collection_dimension_count", 2 if adjusted_dimension_count in (3, 4) else 0)
    datum_dimension_count = data_element.get("datum_dimension_count", adjusted_dimension_count - collection_dimension_count)
    data_descriptor = DataAndMetadata.DataDescriptor(is_sequence, collection_dimension_count, datum_dimension_count)

    # dimensional calibrations
    dimensional_calibrations = None
    if "spatial_calibrations" in data_element:
        dimensional_calibrations_list = typing.cast(typing.List[typing.Any], data_element.get("spatial_calibrations"))
        if len(dimensional_calibrations_list) == dimension_count:
            dimensional_calibrations = list()
            for dimension_calibration in dimensional_calibrations_list:
                offset = float(dimension_calibration.get("offset", 0.0))
                scale = float(dimension_calibration.get("scale", 1.0))
                units = dimension_calibration.get("units", "")
                units = str(units) if units is not None else str()
                if scale != 0.0:
                    dimensional_calibrations.append(Calibration.Calibration(offset, scale, units))
                else:
                    dimensional_calibrations.append(Calibration.Calibration())

    # intensity calibration
    intensity_calibration = None
    if "intensity_calibration" in data_element:
        intensity_calibration_dict = typing.cast(typing.Dict[str, typing.Any], data_element.get("intensity_calibration"))
        offset = float(intensity_calibration_dict.get("offset", 0.0))
        scale = float(intensity_calibration_dict.get("scale", 1.0))
        units = intensity_calibration_dict.get("units", "")
        units = str(units) if units is not None else str()
        if scale != 0.0:
            intensity_calibration = Calibration.Calibration(offset, scale, units)

    # properties (general tags)
    metadata: typing.Dict[str, typing.Any] = dict()
    if "metadata" in data_element:
        metadata.update(Utility.clean_dict(typing.cast(typing.Dict[str, typing.Any], data_element.get("metadata"))))
    if "properties" in data_element and data_element["properties"]:
        hardware_source_metadata = metadata.setdefault("hardware_source", dict())
        hardware_source_metadata.update(Utility.clean_dict(typing.cast(typing.Dict[str, typing.Any], data_element.get("properties"))))

    # dates are _local_ time and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
    # time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
    # daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
    # timezone is for conversion and is the Olson timezone string.
    # datetime.datetime.strptime(datetime.datetime.isoformat(datetime.datetime.now()), "%Y-%m-%dT%H:%M:%S.%f" )
    # datetime_modified, datetime_modified_tz, datetime_modified_dst, datetime_modified_tzname is the time at which this image was modified.
    # datetime_original, datetime_original_tz, datetime_original_dst, datetime_original_tzname is the time at which this image was created.
    timestamp = data_element.get("timestamp", DateTime.utcnow())
    datetime_item = data_element.get("datetime_modified", Utility.get_datetime_item_from_utc_datetime(timestamp))

    local_datetime = Utility.get_datetime_from_datetime_item(datetime_item)
    assert local_datetime
    dst_value = datetime_item.get("dst", "+00")
    tz_value = datetime_item.get("tz", "+0000")
    timezone = datetime_item.get("timezone")
    time_zone = { "dst": dst_value, "tz": tz_value}
    if timezone is not None:
        time_zone["timezone"] = timezone
    # note: dst is informational only; tz already include dst
    tz_adjust = (int(tz_value[1:3]) * 60 + int(tz_value[3:5])) * (-1 if tz_value[0] == '-' else 1)
    utc_datetime = local_datetime - datetime.timedelta(minutes=tz_adjust)  # tz_adjust already contains dst_adjust
    timestamp = utc_datetime

    return DataAndMetadata.new_data_and_metadata(data,
                                                 intensity_calibration=intensity_calibration,
                                                 dimensional_calibrations=dimensional_calibrations,
                                                 metadata=metadata,
                                                 timestamp=timestamp,
                                                 data_descriptor=data_descriptor,
                                                 timezone=timezone,
                                                 timezone_offset=tz_value)


def create_data_element_from_data_item(data_item: DataItem.DataItem, include_data: bool = True) -> DataElementType:
    data_element: DataElementType = dict()
    data_element["version"] = 1
    data_element["reader_version"] = 1
    if data_item.has_data:
        data_element["large_format"] = bool(data_item.large_format)
        if include_data:
            data_element["data"] = data_item.data
        dimensional_calibrations = data_item.dimensional_calibrations
        if dimensional_calibrations is not None:
            calibrations_element = list()
            for calibration in dimensional_calibrations:
                calibration_element = { "offset": calibration.offset, "scale": calibration.scale, "units": calibration.units }
                calibrations_element.append(calibration_element)
            data_element["spatial_calibrations"] = calibrations_element
        intensity_calibration = data_item.intensity_calibration
        if intensity_calibration is not None:
            intensity_calibration_element = { "offset": intensity_calibration.offset, "scale": intensity_calibration.scale, "units": intensity_calibration.units }
            data_element["intensity_calibration"] = intensity_calibration_element
        if data_item.is_sequence:
            data_element["is_sequence"] = data_item.is_sequence
        data_element["collection_dimension_count"] = data_item.collection_dimension_count
        data_element["datum_dimension_count"] = data_item.datum_dimension_count
        data_item_metadata = data_item.metadata or dict()
        data_element["metadata"] = copy.deepcopy(data_item_metadata)
        data_element["properties"] = copy.deepcopy(data_item_metadata.get("hardware_source", dict()))
        data_element["title"] = data_item.title
        data_element["source_file_path"] = data_item.source_file_path.as_posix() if data_item.source_file_path else None
        tz_value = data_item.timezone_offset
        timezone = data_item.timezone
        dst_minutes = None
        time_zone_dict = data_item_metadata.get("description", dict()).get("time_zone")
        if time_zone_dict:
            # note: dst is informational only; tz already include dst
            if tz_value is None:
                tz_value = time_zone_dict["tz"]
            dst_minutes = int(time_zone_dict["dst"])
            if timezone is None:
                timezone = time_zone_dict.get("timezone")
        tz_minutes: typing.Optional[int]
        if tz_value is not None:
            tz_minutes = (int(tz_value[1:3]) * 60 + int(tz_value[3:5])) * (-1 if tz_value[0] == '-' else 1)
        else:
            tz_minutes = None
        data_element["datetime_modified"] = Utility.get_datetime_item_from_utc_datetime(data_item.created, tz_minutes, dst_minutes, timezone)
        data_element["datetime_original"] = Utility.get_datetime_item_from_utc_datetime(data_item.created, tz_minutes, dst_minutes, timezone)
        data_element["uuid"] = str(data_item.uuid)
        # operation
        # graphics
    return data_element


def create_data_element_from_extended_data(xdata: DataAndMetadata.DataAndMetadata) -> DataElementType:
    data_element: DataElementType = dict()
    data_element["version"] = 1
    data_element["reader_version"] = 1
    data_element["data"] = xdata.data
    data_element["large_format"] = len(xdata.dimensional_shape) > 2
    dimensional_calibrations = xdata.dimensional_calibrations
    calibrations_element = list()
    for calibration in dimensional_calibrations:
        calibration_element = { "offset": calibration.offset, "scale": calibration.scale, "units": calibration.units }
        calibrations_element.append(calibration_element)
    data_element["spatial_calibrations"] = calibrations_element
    intensity_calibration = xdata.intensity_calibration
    intensity_calibration_element = { "offset": intensity_calibration.offset, "scale": intensity_calibration.scale, "units": intensity_calibration.units }
    data_element["intensity_calibration"] = intensity_calibration_element
    if xdata.is_sequence:
        data_element["is_sequence"] = xdata.is_sequence
    data_element["collection_dimension_count"] = xdata.collection_dimension_count
    data_element["datum_dimension_count"] = xdata.datum_dimension_count
    data_element["metadata"] = xdata.metadata
    # properties are redundant; but here for backwards compatibility
    data_element["properties"] = xdata.metadata.get("hardware_source", dict())
    tz_minutes = None
    dst_minutes = None
    timezone = None
    time_zone_dict = xdata.metadata.get("description", dict()).get("time_zone")
    if time_zone_dict:
        # note: dst is informational only; tz already include dst
        tz_value = time_zone_dict["tz"]
        tz_minutes = (int(tz_value[1:3]) * 60 + int(tz_value[3:5])) * (-1 if tz_value[0] == '-' else 1)
        dst_minutes = int(time_zone_dict["dst"])
        timezone = time_zone_dict.get("timezone")
    data_element["datetime_modified"] = Utility.get_datetime_item_from_utc_datetime(xdata.timestamp, tz_minutes, dst_minutes, timezone)
    data_element["datetime_original"] = Utility.get_datetime_item_from_utc_datetime(xdata.timestamp, tz_minutes, dst_minutes, timezone)
    return data_element


# return True if data is grayscale.
def is_grayscale(data: _DataArrayType) -> bool:
    if Image.is_data_rgb(data) or Image.is_data_rgba(data):
        return numpy.array_equal(data[..., 0],data[..., 1]) and numpy.array_equal(data[..., 1],data[..., 2])
    return True


# based roughly on https://github.com/imageio/imageio/blob/896aa50c1797c3642149aa3110547d85190017db/imageio/core/util.py#L45
def convert_to_uint8(d: _DataArrayType) -> numpy.typing.NDArray[numpy.uint8]:
    if d.dtype == numpy.dtype(numpy.uint8):
        return d
    # check the int types first.
    if d.dtype == numpy.dtype(numpy.uint16):
        return typing.cast(numpy.typing.NDArray[numpy.uint8], numpy.right_shift(d, 8).astype(numpy.uint8))
    elif d.dtype == numpy.dtype(numpy.uint32):
        return typing.cast(numpy.typing.NDArray[numpy.uint8], numpy.right_shift(d, 24).astype(numpy.uint8))
    elif d.dtype == numpy.dtype(numpy.uint64):
        return typing.cast(numpy.typing.NDArray[numpy.uint8], numpy.right_shift(d, 56).astype(numpy.uint8))
    # now check float type
    mn, mx = numpy.nanmin(d), numpy.nanmax(d)
    if str(d.dtype).startswith("float") and mn >= 0 and mx <= 1:
        return typing.cast(numpy.typing.NDArray[numpy.uint8], numpy.round(d.astype(numpy.float64) * 255).astype(numpy.uint8))
    elif numpy.isfinite(mn) and numpy.isfinite(mx) and mn != mx:
        return typing.cast(numpy.typing.NDArray[numpy.uint8], numpy.round((d.astype(d.astype(numpy.float64)) - mn) / (mx - mn) * 255).astype(numpy.uint8))
    else:
        return d.astype(numpy.uint8)

# read image file. convert to dtype if it is grayscale.
def read_image_from_file(filename: pathlib.Path) -> _DataArrayType:
    if str(filename).startswith(":"):
        return numpy.zeros((20, 20, 4), numpy.uint8)
    # TODO: fix typing when imageio gets their numpy typing correct.
    image = imageio.imread(filename, index=0)
    if image is not None:
        image_u8 = convert_to_uint8(image)
        if len(image_u8.shape) == 3:
            rgba_image: numpy.typing.NDArray[numpy.uint8]
            if image_u8.shape[-1] == 3:
                rgba_image = numpy.empty(image_u8.shape[:-1] + (4,), numpy.uint8)
                rgba_image[..., 0] = image_u8[..., 2]
                rgba_image[..., 1] = image_u8[..., 1]
                rgba_image[..., 2] = image_u8[..., 0]
                rgba_image[..., 3] = 255
            else:
                assert image_u8.shape[-1] == 4
                rgba_image = numpy.empty(image_u8.shape[:-1] + (4,), numpy.uint8)
                rgba_image[..., 0] = image_u8[..., 2]
                rgba_image[..., 1] = image_u8[..., 1]
                rgba_image[..., 2] = image_u8[..., 0]
                rgba_image[..., 3] = image_u8[..., 3]
            if is_grayscale(rgba_image):
                rgba_image = Image.convert_to_grayscale(rgba_image)
        else:
            assert len(image_u8.shape) == 2
            rgba_image = image_u8
        assert rgba_image is not None
        return rgba_image
    raise IOError()


class StandardImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id: str, name: str, extensions: typing.Sequence[str]) -> None:
        super().__init__(io_handler_id, name, extensions)

    def read_data_elements(self, extension: str, path: pathlib.Path) -> typing.List[DataElementType]:
        data = None
        try:
            data = read_image_from_file(path)
        except Exception as e:
            print(e)
        if data is not None:
            data_element: DataElementType = dict()
            data_element["version"] = 1
            data_element["data"] = data
            if path.exists():
                try:
                    file_datetime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
                    data_element["datetime_modified"] = Utility.get_datetime_item_from_datetime(file_datetime)
                except Exception:
                    pass
            return [data_element]
        return list()

    def can_write(self, data_metadata: DataAndMetadata.DataMetadata, extension: str) -> bool:
        return len(data_metadata.dimensional_shape) == 2

    def write_display_item(self, display_item: DisplayItem.DisplayItem, path: pathlib.Path, extension: str) -> None:
        display_data_channel = display_item.display_data_channel
        assert display_data_channel
        display_values = display_data_channel.get_latest_computed_display_values()
        assert display_values
        data = display_values.display_rgba  # export the display rather than the data for these types
        assert data is not None
        # TODO: fix typing when imageio gets their numpy typing correct.
        imageio.imwrite(path, Image.get_rgb_view(data), extension="." + extension)


class CSVImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id: str, name: str, extensions: typing.Sequence[str]) -> None:
        super().__init__(io_handler_id, name, extensions)

    def read_data_elements(self, extension: str, path: pathlib.Path) -> typing.List[DataElementType]:
        data = numpy.loadtxt(str(path), delimiter=',')
        if data is not None:
            data_element: DataElementType = dict()
            data_element["data"] = data
            return [data_element]
        return list()

    def can_write(self, data_metadata: DataAndMetadata.DataMetadata, extension: str) -> bool:
        return True

    def write_display_item(self, display_item: DisplayItem.DisplayItem, path: pathlib.Path, extension: str) -> None:
        data_item = display_item.data_item
        assert data_item
        data = data_item.data
        if data is not None:
            numpy.savetxt(path, data, delimiter=', ')


def build_table(display_item: DisplayItem.DisplayItem) -> typing.Tuple[typing.List[str], typing.List[_DataArrayType]]:
    data_items = display_item.data_items
    assert all([data_item.is_data_1d for data_item in data_items])

    def make_x_data(calibration: Calibration.Calibration, length: int) -> _DataArrayType:
        return numpy.linspace(calibration.offset, calibration.offset + (length - 1) * calibration.scale, length)

    xdata0 = data_items[0].xdata
    calibration0 = xdata0.dimensional_calibrations[0] if xdata0 else Calibration.Calibration()
    if all([calibration0 == (data_item.xdata.dimensional_calibrations[0] if data_item.xdata else Calibration.Calibration()) for data_item in data_items]):
        length = max([data_item.xdata.data_shape[0] if data_item.xdata else 0 for data_item in data_items])
        data_list = [make_x_data(calibration0, length)]
        headers = [f"X ({calibration0.units or 'pixel'})"]
        for index in range(len(display_item.display_layers)):
            display_data_channel = display_item.get_display_layer_display_data_channel(index)
            assert display_data_channel
            display_values = display_data_channel.get_latest_computed_display_values()
            assert display_values
            xdata = display_values.display_data_and_metadata
            assert xdata
            data = xdata.data
            assert data is not None
            data_list.append(xdata.intensity_calibration.convert_array_to_calibrated_value(data))
            label = display_item.get_display_layer_property(index, "label") or f"Data {index}"
            label = label + f" ({xdata.intensity_calibration.units or 'None'})"
            headers.append(label)
    else:
        data_list = list()
        headers = list()
        for index in range(len(display_item.display_layers)):
            display_data_channel = display_item.get_display_layer_display_data_channel(index)
            assert display_data_channel
            display_values = display_data_channel.get_latest_computed_display_values()
            assert display_values
            xdata = display_values.display_data_and_metadata
            assert xdata
            data = xdata.data
            assert data is not None
            data_list.append(make_x_data(xdata.dimensional_calibrations[0], xdata.data_shape[0]))
            data_list.append(xdata.intensity_calibration.convert_array_to_calibrated_value(data))
            label = display_item.get_display_layer_property(index, "label") or f"Data {index}"
            x_label = "X " + label + f" ({xdata.dimensional_calibrations[0].units or 'pixel'})"
            y_label = "Y " + label + f" ({xdata.intensity_calibration.units or 'None'})"
            headers.append(x_label)
            headers.append(y_label)

    return headers, data_list


class CSV1ImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id: str, name: str, extensions: typing.Sequence[str]) -> None:
        super().__init__(io_handler_id, name, extensions)
        self.supports_composite_data = True

    def read_data_elements(self, extension: str, path: pathlib.Path) -> typing.List[DataElementType]:
        return list()

    def can_write(self, data_metadata: DataAndMetadata.DataMetadata, extension: str) -> bool:
        return data_metadata.is_data_1d

    def write_display_item(self, display_item: DisplayItem.DisplayItem, path: pathlib.Path, extension: str) -> None:
        headers, data_list = build_table(display_item)

        newline = "\n"
        delimiter = ", "
        format_ = "{}" # We need to stick with an empty format string here, otherwise we will have problems with the
                       # fill value of zip_longest
        row_template = delimiter.join([format_] * len(data_list))
        row_template += newline
        header_template = "# " + row_template

        with open(path, "w+") as f:
            f.write(header_template.format(*headers))
            for row in itertools.zip_longest(*data_list, fillvalue=""):
                f.write(row_template.format(*row))


class NDataImportExportHandler(ImportExportHandler):

    def __init__(self, io_handler_id: str, name: str, extensions: typing.Sequence[str]) -> None:
        super().__init__(io_handler_id, name, extensions)

    def read_data_elements(self, extension: str, path: pathlib.Path) -> typing.List[DataElementType]:
        zip_file = zipfile.ZipFile(path, 'r')
        namelist = zip_file.namelist()
        if "metadata.json" in namelist and "data.npy" in namelist:
            metadata = json.loads(zip_file.read("metadata.json").decode("utf-8"))
            data_bytes = zip_file.read("data.npy")
            data = numpy.load(io.BytesIO(data_bytes))
            if data is not None:
                data_element = metadata
                data_element["data"] = data
                return [data_element]
        return list()

    def can_write(self, data_metadata: DataAndMetadata.DataMetadata, extension: str) -> bool:
        return True

    def write_display_item(self, display_item: DisplayItem.DisplayItem, path: pathlib.Path, extension: str) -> None:
        data_item = display_item.data_item
        assert data_item
        data_element = create_data_element_from_data_item(data_item, include_data=False)
        data = data_item.data
        if data is not None:
            root = str(path.parent)
            metadata_path = root + "_metadata.json"
            data_path = root + "_data.npy"
            try:
                with open(metadata_path, "w") as fp:
                    json.dump(data_element, fp)
                numpy.save(data_path, data)
                zip_file = zipfile.ZipFile(path, 'w')
                zip_file.write(metadata_path, "metadata.json")
                zip_file.write(data_path, "data.npy")
            finally:
                os.remove(metadata_path)
                os.remove(data_path)


class NumPyImportExportHandler(ImportExportHandler):
    """A file import/export handler to read/write the npy file type.

    The npy file type is the one included in the NumPy package.

    This i/o handler will optionally read metadata with the same name but the
    '.json' suffix located in the same directory.

    This i/o handler will write metadata to a file with the same name in the same
    directory but with the '.json' suffix.
    """

    def __init__(self, io_handler_id: str, name: str, extensions: typing.Sequence[str]) -> None:
        super().__init__(io_handler_id, name, extensions)

    def read_data_elements(self, extension: str, path: pathlib.Path) -> typing.List[DataElementType]:
        data = numpy.load(str(path))
        metadata_path = path.with_suffix(".json")
        if metadata_path.exists():
            with open(metadata_path) as f:
                metadata = json.load(f)
        else:
            metadata = dict()
        if data is not None:
            data_element = metadata
            data_element["data"] = data
            return [data_element]
        return list()

    def can_write(self, data_metadata: DataAndMetadata.DataMetadata, extension: str) -> bool:
        return True

    def write_display_item(self, display_item: DisplayItem.DisplayItem, path: pathlib.Path, extension: str) -> None:
        data_item = display_item.data_item
        assert data_item
        data_path = path
        metadata_path = data_path.with_suffix(".json")
        data_element = create_data_element_from_data_item(data_item, include_data=False)
        data = data_item.data
        if data is not None:
            try:
                with open(str(metadata_path), "w") as fp:
                    json.dump(data_element, fp)
                numpy.save(str(data_path), data)
            except Exception:
                os.remove(str(metadata_path))
                os.remove(str(data_path))
                raise


# Register the intrinsic I/O handlers.
ImportExportManager().register_io_handler(StandardImportExportHandler("jpeg-io-handler", "JPEG", ["jpg", "jpeg"]))
ImportExportManager().register_io_handler(StandardImportExportHandler("png-io-handler", "PNG", ["png"]))
ImportExportManager().register_io_handler(StandardImportExportHandler("gif-io-handler", "GIF", ["gif"]))
ImportExportManager().register_io_handler(StandardImportExportHandler("bmp-io-handler", "BMP", ["bmp"]))
# ImportExportManager().register_io_handler(StandardImportExportHandler("tiff-io-handler", "TIFF", ["tif", "tiff"]))
# ImportExportManager().register_io_handler(StandardImportExportHandler("cr2-io-handler", "CR2", ["cr2"]))
# ImportExportManager().register_io_handler(StandardImportExportHandler("nef-io-handler", "NEF", ["nef"]))
ImportExportManager().register_io_handler(CSVImportExportHandler("csv-io-handler", "CSV Raw", ["csv"]))
ImportExportManager().register_io_handler(CSV1ImportExportHandler("csv1-io-handler", "CSV 1D", ["csv"]))
ImportExportManager().register_io_handler(NDataImportExportHandler("ndata1-io-handler", "NData 1", ["ndata1"]))
ImportExportManager().register_io_handler(NumPyImportExportHandler("numpy-io-handler", "Raw NumPy", ["npy"]))
ImportExportManager().register_io_handler(StandardImportExportHandler("webp-io-handler", "WebP", ["webp"]))



# standard libraries
import contextlib
import copy
import math
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift import Facade
from nion.swift.model import DataItem
from nion.swift.model import DataStructure
from nion.swift.model import Graphics
from nion.swift.model import Schema
from nion.swift.model import Symbolic
from nion.swift.model import Utility
from nion.swift.test import TestContext
from nion.ui import TestUI


Facade.initialize()


class TestDataStructureClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        TestContext.end_leaks(self)

    def test_changing_entity_field_updates_data_structure(self):
        TestModel = Schema.entity("test_model", None, None, {"flag": Schema.prop(Schema.BOOLEAN)})
        with TestContext.create_memory_context() as test_context:
            DataStructure.DataStructure.register_entity(TestModel)
            try:
                document_model = test_context.create_document_model()
                data_structure = DataStructure.DataStructure(structure_type="test_model")
                document_model.append_data_structure(data_structure)
                self.assertIsNotNone(data_structure.entity)
                data_structure.entity._set_field_value("flag", False)
                self.assertEqual(data_structure.flag, False)
                self.assertEqual(data_structure.flag, data_structure.entity._get_field_value("flag"))
                data_structure.entity._set_field_value("flag", True)
                self.assertEqual(data_structure.flag, True)
                self.assertEqual(data_structure.flag, data_structure.entity._get_field_value("flag"))
            finally:
                DataStructure.DataStructure.unregister_entity(TestModel)



"""
    Contains classes related to thumbnail generation.
"""

from __future__ import annotations

# standard libraries
import functools
import threading
import typing
import uuid
import weakref

# third-party libraries
import numpy

# local libraries
from nion.swift import DisplayPanel
from nion.swift.model import Utility
from nion.swift.model import DisplayItem
from nion.ui import DrawingContext
from nion.ui import UserInterface
from nion.utils import Event
from nion.utils import ReferenceCounting
from nion.utils import ThreadPool

_NDArray = numpy.typing.NDArray[typing.Any]
_ThumbnailSourceWeakRef = typing.Callable[[], typing.Optional["ThumbnailSource"]]  # Python 3.9+


class ThumbnailProcessor:
    """Processes thumbnails for a display in a thread."""

    def __init__(self, display_item: DisplayItem.DisplayItem):
        self.__display_item = display_item
        self.__recompute_lock = threading.RLock()
        self.__dispatcher = ThreadPool.SingleItemDispatcher(minimum_period=0.5)
        self.__display_item_about_to_close_listener = self.__display_item.about_to_close_event.listen(self.__about_to_close_display_item)
        self.__cache = self.__display_item._display_cache
        self.__cache_property_name = "thumbnail_data"
        self.width = 256
        self.height = 256
        self.on_thumbnail_updated: typing.Optional[typing.Callable[[], None]] = None

    def close(self) -> None:
        self.on_thumbnail_updated = None
        self.__dispatcher.close()
        self.__dispatcher = typing.cast(typing.Any, None)
        self.__display_item = typing.cast(typing.Any, None)
        self.__display_item_about_to_close_listener.close()
        self.__display_item_about_to_close_listener = typing.cast(typing.Any, None)

    def __about_to_close_display_item(self) -> None:
        self.close()

    # used internally and for testing
    @property
    def _is_cached_value_dirty(self) -> bool:
        return self.__cache.is_cached_value_dirty(self.__display_item, self.__cache_property_name)

    # thread safe
    def mark_data_dirty(self) -> None:
        """ Called from item to indicate its data or metadata has changed."""
        self.__cache.set_cached_value_dirty(self.__display_item, self.__cache_property_name)

    def __get_cached_value(self) -> typing.Optional[_NDArray]:
        return typing.cast(typing.Optional[_NDArray], self.__cache.get_cached_value(self.__display_item, self.__cache_property_name))

    def get_cached_data(self) -> typing.Optional[_NDArray]:
        """Return the cached data for this processor.

        This method is thread safe and always returns quickly, using the cached data.
        """
        return self.__get_cached_value()

    def __get_calculated_data(self, ui: UserInterface.UserInterface) -> typing.Optional[DrawingContext.RGBA32Type]:
        drawing_context, shape = DisplayPanel.preview(DisplayPanel.DisplayPanelUISettings(ui), self.__display_item, 512, 512)
        thumbnail_drawing_context = DrawingContext.DrawingContext()
        thumbnail_drawing_context.scale(self.width / 512, self.height / 512)
        thumbnail_drawing_context.translate(0, (shape[1] - shape[0]) * 0.5)
        thumbnail_drawing_context.add(drawing_context)
        return ui.create_rgba_image(thumbnail_drawing_context, self.width, self.height)

    def recompute(self, ui: UserInterface.UserInterface) -> None:
        self.__dispatcher.dispatch(functools.partial(self.recompute_data, ui))

    def recompute_data(self, ui: UserInterface.UserInterface) -> None:
        """Compute the data associated with this processor.

        This method is thread safe and may take a long time to return. It should not be called from
         the UI thread. Upon return, the results will be calculated with the latest data available
         and the cache will not be marked dirty.
        """
        with self.__recompute_lock:
            try:
                calculated_data = self.__get_calculated_data(ui)
            except Exception as e:
                import traceback
                traceback.print_exc()
                traceback.print_stack()
                raise
            if calculated_data is None:
                calculated_data = numpy.zeros((self.height, self.width), dtype=numpy.uint32)
            self.__cache.set_cached_value(self.__display_item, self.__cache_property_name, calculated_data)
        if callable(self.on_thumbnail_updated):
            self.on_thumbnail_updated()


class ThumbnailSource(ReferenceCounting.ReferenceCounted):
    """Produce a thumbnail for a display."""

    def __init__(self, ui: UserInterface.UserInterface, display_item: DisplayItem.DisplayItem) -> None:
        super().__init__()
        self._ui = ui
        self._display_item = display_item

        self.thumbnail_updated_event = Event.Event()
        self.__thumbnail_processor = ThumbnailProcessor(display_item)

        def thumbnail_changed() -> None:
            thumbnail_processor = self.__thumbnail_processor
            if thumbnail_processor:
                thumbnail_processor.mark_data_dirty()
                thumbnail_processor.recompute(ui)

        self.__display_changed_event_listener = display_item.display_changed_event.listen(thumbnail_changed)

        def thumbnail_updated() -> None:
            self.thumbnail_updated_event.fire()

        self.__thumbnail_processor.on_thumbnail_updated = thumbnail_updated

        # initial recompute, if required
        if self.__thumbnail_processor._is_cached_value_dirty:
            self.__thumbnail_processor.recompute(ui)

        def display_item_will_close() -> None:
            if self.__thumbnail_processor:
                self.__thumbnail_processor.close()
                self.__thumbnail_processor = typing.cast(typing.Any, None)

        self.__display_will_close_listener = display_item.about_to_be_removed_event.listen(display_item_will_close)

    def about_to_delete(self) -> None:
        self.__display_will_close_listener.close()
        self.__display_will_close_listener = typing.cast(typing.Any, None)
        if self.__thumbnail_processor:
            self.__thumbnail_processor.close()
            self.__thumbnail_processor = typing.cast(typing.Any, None)
        self.__display_changed_event_listener.close()
        self.__display_changed_event_listener = typing.cast(typing.Any, None)
        super().about_to_delete()

    def add_ref(self) -> ThumbnailSource:
        super().add_ref()
        return self

    @property
    def thumbnail_data(self) -> typing.Optional[_NDArray]:
        return self.__thumbnail_processor.get_cached_data() if self.__thumbnail_processor else None

    def recompute_data(self) -> None:
        self.__thumbnail_processor.recompute_data(self._ui)

    # used for testing
    @property
    def _is_thumbnail_dirty(self) -> bool:
        return self.__thumbnail_processor._is_cached_value_dirty


class ThumbnailManager(metaclass=Utility.Singleton):
    """Manages thumbnail sources for displays."""

    def __init__(self) -> None:
        self.__thumbnail_sources: typing.Dict[uuid.UUID, _ThumbnailSourceWeakRef] = dict()
        self.__lock = threading.RLock()

    def thumbnail_sources(self) -> typing.Dict[uuid.UUID, _ThumbnailSourceWeakRef]:
        return self.__thumbnail_sources

    def thumbnail_source_for_display_item(self, ui: UserInterface.UserInterface, display_item: DisplayItem.DisplayItem) -> ThumbnailSource:
        """Returned ThumbnailSource must be closed."""
        with self.__lock:
            thumbnail_source_ref = self.__thumbnail_sources.get(display_item.uuid)
            thumbnail_source = thumbnail_source_ref() if thumbnail_source_ref else None
            if not thumbnail_source:
                thumbnail_source = ThumbnailSource(ui, display_item)
                self.__thumbnail_sources[display_item.uuid] = weakref.ref(thumbnail_source)
                weakref.finalize(thumbnail_source, self.__thumbnail_sources.pop, display_item.uuid)
            else:
                assert thumbnail_source._ui == ui
            return thumbnail_source

    def thumbnail_data_for_display_item(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> typing.Optional[_NDArray]:
        with self.__lock:
            thumbnail_source_ref = self.__thumbnail_sources.get(display_item.uuid) if display_item else None
            thumbnail_source = thumbnail_source_ref() if thumbnail_source_ref else None
            if thumbnail_source:
                return thumbnail_source.thumbnail_data
            return None



from __future__ import annotations

# standard libraries
import asyncio
import collections
import contextlib
import dataclasses
import datetime
import functools
import logging
import os
import pathlib
import sys
import threading
import time
import traceback
import types

# third party libraries
import typing

import numpy

# local libraries
from nion.utils import DateTime


# datetimes are _local_ datetimes and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
# time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
# daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
# timezone is for conversion and is the Olson timezone string.
# datetime_item is a dictionary with entries for the local_datetime, tz (timezone offset), and
# dst (daylight savings time offset). it may optionally include tz_name (timezone name), if available.
def get_datetime_item_from_datetime(datetime_local: datetime.datetime, tz_minutes: typing.Optional[int] = None,
                                    dst_minutes: typing.Optional[int] = None, timezone: typing.Optional[str] = None) -> typing.Dict[str, typing.Any]:
    # dst is information, tz already includes dst
    datetime_item = dict()
    datetime_item["local_datetime"] = datetime_local.isoformat()
    if tz_minutes is None:
        tz_minutes = int(round((datetime.datetime.now() - DateTime.utcnow()).total_seconds())) // 60
    if dst_minutes is None:
        dst_minutes = 60 if time.localtime().tm_isdst else 0
    datetime_item["tz"] = '{0:+03d}{1:02d}'.format(tz_minutes // 60, tz_minutes % 60)
    datetime_item["dst"] = "+{0:02d}".format(dst_minutes)
    if timezone is not None:
        datetime_item["timezone"] = timezone
    return datetime_item


def get_current_datetime_item() -> typing.Dict[str, typing.Any]:
    return get_datetime_item_from_datetime(datetime.datetime.now())


def get_datetime_item_from_utc_datetime(datetime_utc: datetime.datetime, tz_minutes: typing.Optional[int] = None,
                                        dst_minutes: typing.Optional[int] = None,
                                        timezone: typing.Optional[str] = None) -> typing.Dict[str, typing.Any]:
    # dst is information, tz already includes dst
    if tz_minutes is None:
        tz_minutes = int(round((datetime.datetime.now() - DateTime.utcnow()).total_seconds())) // 60
    return get_datetime_item_from_datetime(datetime_utc + datetime.timedelta(minutes=tz_minutes), tz_minutes,
                                           dst_minutes, timezone)


local_utcoffset_override: typing.Optional[typing.List[int]] = None  # for testing
def local_utcoffset_minutes(datetime_local: typing.Optional[datetime.datetime] = None) -> int:
    if local_utcoffset_override is not None:
        return local_utcoffset_override[0]
    now = datetime.datetime.now()
    return int((now - now.astimezone(tz=datetime.timezone.utc).replace(tzinfo=None)).total_seconds() / 60)


class TimezoneMinutesToStringConverter:
    def convert(self, value: typing.Optional[int]) -> typing.Optional[str]:
        return "{0:+03d}{1:02d}".format(value // 60, value % 60) if value is not None else None

    def convert_back(self, value: typing.Optional[str]) -> typing.Optional[int]:
        return (int(value[1:3]) * 60 + int(value[3:5])) * (-1 if value[0] == '-' else 1) if value is not None else None


local_timezone_override: typing.Optional[typing.List[str]] = None  # for testing
try:
    import tzlocal

    def get_local_timezone() -> typing.Optional[str]:
        if local_timezone_override is None:
            # see note https://github.com/regebro/tzlocal/issues/117#issuecomment-939351032
            return str(tzlocal.get_localzone())
        else:
            return local_timezone_override[0]
except ImportError:
    def get_local_timezone() -> typing.Optional[str]:
        return None if local_timezone_override is None else local_timezone_override[0]


# return python datetime object from a datetime_item. may return None if the datetime element is
# not properly formatted.
def get_datetime_from_datetime_item(datetime_item: typing.Dict[str, typing.Any]) -> typing.Optional[datetime.datetime]:
    local_datetime = datetime_item.get("local_datetime", str())
    if len(local_datetime) == 26:
        return datetime.datetime.strptime(local_datetime, "%Y-%m-%dT%H:%M:%S.%f")
    elif len(local_datetime) == 19:
        return datetime.datetime.strptime(local_datetime, "%Y-%m-%dT%H:%M:%S")
    return None


class Singleton(type):
    def __init__(cls, name: str, bases: typing.Tuple[type], dict: typing.Dict[str, typing.Any]):
        super(Singleton, cls).__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args: typing.List[typing.Any], **kw: typing.Dict[str, typing.Any]) -> typing.Any:
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


DirtyValue = typing.Any
CleanValue = typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any], typing.Tuple[typing.Any], str, float, int, bool, None]


def clean_dict(d0: typing.Dict[str, DirtyValue], clean_item_fn: typing.Optional[typing.Callable[[DirtyValue], CleanValue]] = None) -> typing.Dict[str, CleanValue]:
    """Return a json-clean dict. Will log info message for failures."""
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    d: typing.Dict[str, CleanValue] = dict()
    for key in d0:
        cleaned_item = clean_item_fn(d0[key])
        if cleaned_item is not None:
            d[key] = cleaned_item
    return d


def clean_list(l0: typing.List[DirtyValue], clean_item_fn: typing.Optional[typing.Callable[[DirtyValue], CleanValue]] = None) -> typing.List[CleanValue]:
    """Return a json-clean list. Will log info message for failures."""
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    l: typing.List[CleanValue] = list()
    for index, item in enumerate(l0):
        cleaned_item = clean_item_fn(item)
        l.append(cleaned_item)
    return l


def clean_tuple(t0: typing.Tuple[DirtyValue], clean_item_fn: typing.Optional[typing.Callable[[DirtyValue], CleanValue]] = None) -> typing.Tuple[CleanValue]:
    """Return a json-clean tuple. Will log info message for failures."""
    clean_item_fn = clean_item_fn if clean_item_fn else clean_item
    l: typing.List[CleanValue] = list()
    for index, item in enumerate(t0):
        cleaned_item = clean_item_fn(item)
        l.append(cleaned_item)
    return typing.cast(typing.Tuple[CleanValue], tuple(l))


type_lookup: typing.Mapping[typing.Type[typing.Any], typing.Callable[[DirtyValue], CleanValue]] = {
    dict: clean_dict,
    list: clean_list,
    tuple: clean_tuple,
    numpy.float32: lambda x: float(x),
    numpy.float64: lambda x: float(x),
    numpy.int16: lambda x: int(x),
    numpy.uint16: lambda x: int(x),
    numpy.int32: lambda x: int(x),
    numpy.uint32: lambda x: int(x),
    numpy.int64: lambda x: int(x),
    numpy.uint64: lambda x: int(x),
    float: lambda x: float(x),
    str: lambda x: str(x),
    int: lambda x: int(x),
    numpy.bool_: lambda x: bool(x),
    bool: lambda x: bool(x),
    type(None): lambda x: None
}


type_lookup_no_list: typing.Mapping[typing.Type[typing.Any], typing.Callable[[DirtyValue], CleanValue]] = {
    dict: clean_dict,
    list: clean_tuple,
    tuple: clean_tuple,
    numpy.float32: lambda x: float(x),
    numpy.float64: lambda x: float(x),
    numpy.int16: lambda x: int(x),
    numpy.uint16: lambda x: int(x),
    numpy.int32: lambda x: int(x),
    numpy.uint32: lambda x: int(x),
    numpy.int64: lambda x: int(x),
    numpy.uint64: lambda x: int(x),
    float: lambda x: float(x),
    str: lambda x: str(x),
    int: lambda x: int(x),
    numpy.bool_: lambda x: bool(x),
    bool: lambda x: bool(x),
    type(None): lambda x: None
}


def clean_item(i: DirtyValue) -> CleanValue:
    """Return a json-clean item or None. Will log info message for failure."""
    itype = type(i)
    c = type_lookup.get(itype, None)
    if c:
        return c(i)
    else:
        logging.info("[1] Unable to handle type %s", itype)
        import traceback
        traceback.print_stack()
        return None


def clean_item_no_list(i: DirtyValue) -> CleanValue:
    """Return a json-clean item or None. Will log info message for failure."""
    itype = type(i)
    c = type_lookup_no_list.get(itype, None)
    if c:
        return c(i)
    else:
        logging.info("[1] Unable to handle type %s", itype)
        import traceback
        traceback.print_stack()
        return None


def deep_compare_items(item1: CleanValue, item2: CleanValue) -> bool:
    """Deep compare dict/list/value. Return True if equal."""
    item1_type = type(item1)
    if item1_type != type(item2):
        return False
    if item1_type in (dict,):
        item1_dict = typing.cast(typing.Dict[str, CleanValue], item1)
        item2_dict = typing.cast(typing.Dict[str, CleanValue], item2)
        if item1_dict.keys() != item2_dict.keys():
            return False
        for key in item1_dict.keys():
            if not deep_compare_items(item1_dict[key], item2_dict[key]):
                return False
        return True
    elif item1_type in (list, tuple):
        item1_list = typing.cast(typing.List[CleanValue], item1)
        item2_list = typing.cast(typing.List[CleanValue], item2)
        if len(item1_list) != len(item2_list):
            return False
        for child_item1, child_item2 in zip(item1_list, item2_list):
            if not deep_compare_items(child_item1, child_item2):
                return False
        return True
    return item1 == item2


def parse_version(version: str, count: int = 3, max_count: typing.Optional[int] = None) -> typing.List[int]:
    max_count = max_count if max_count is not None else count
    version_components = [int(version_component) for version_component in version.split(".")]
    assert len(version_components) <= max_count
    while len(version_components) < count:
        version_components.append(0)
    return version_components


def compare_versions(version1: str, version2: str) -> int:
    if version1.startswith("~"):
        version1 = version1[1:]
        version_components1 = parse_version(version1, 1, 3)
        assert len(version_components1) > 1
    elif version1 == "1":  # same as "~1.0"
        version1 = "1.0"
        version_components1 = parse_version(version1, 2, 3)
    else:
        version_components1 = parse_version(version1)
    version_components2 = parse_version(version2)
    # print(version_components1, version_components2)
    for version_component1, version_component2 in zip(version_components1, version_components2):
        if version_component1 > version_component2:
            return 1
        elif version_component1 < version_component2:
            return -1
    return 0


class AtomicFileWriter:
    # see https://blog.gocept.com/2013/07/15/reliable-file-updates-with-python/
    # see https://github.com/mahmoud/boltons/blob/1885ed64006982cdd08a70c2cba193e1fe2eb693/boltons/fileutils.py#L343

    def __init__(self, filepath: pathlib.Path):
        self.__filepath = filepath
        self.__temp_filepath = self.__filepath.with_suffix(".temp")
        self.__fp: typing.Optional[typing.TextIO] = None

    def __enter__(self) -> typing.TextIO:
        self.__fp = self.__temp_filepath.open("w")
        return self.__fp

    def __exit__(self, exception_type: typing.Optional[typing.Type[Exception]], value: typing.Optional[Exception], traceback: typing.Optional[types.TracebackType]) -> None:
        assert self.__fp
        self.__fp.flush()
        os.fsync(self.__fp)
        self.__fp.close()
        if exception_type:
            try:
                os.unlink(self.__temp_filepath)
            except Exception:
                pass
        else:
            try:
                os.replace(self.__temp_filepath, self.__filepath)
                if hasattr(os, "O_DIRECTORY"):
                    # ensure the directory has been updated. not available all the time on Windows.
                    dirfd = os.open(os.path.dirname(self.__filepath), os.O_DIRECTORY)
                    os.fsync(dirfd)
                    os.close(dirfd)
            except OSError:
                try:
                    os.unlink(self.__temp_filepath)
                except Exception:
                    pass
                raise


@dataclasses.dataclass
class FPSRecord:
    window_elapsed_times: typing.List[float] = dataclasses.field(default_factory=list)
    window_elapsed_time: float = 0.0
    last_tick: typing.Optional[float] = None
    last_window_frame_mean: float = 0.0
    last_window_frame_std: float = 0.0


def fps_tick(fps_id: str) -> str:
    v = globals().setdefault("__fps_" + fps_id, FPSRecord())
    next_time = time.perf_counter()
    if v.last_tick is not None:
        elapsed_time = next_time - v.last_tick
        v.window_elapsed_times.append(elapsed_time)
        v.window_elapsed_time += elapsed_time
    if v.window_elapsed_time > 1.0:
        v.last_window_frame_mean = numpy.mean(v.window_elapsed_times)
        v.last_window_frame_std = numpy.std(v.window_elapsed_times)
        v.window_elapsed_time = 0.0
        v.window_elapsed_times = list()
    v.last_tick = next_time
    return fps_get(fps_id)


def fps_get(fps_id: str) -> str:
    v = globals().setdefault("__fps_" + fps_id, FPSRecord())
    m = v.last_window_frame_mean
    r = 1 / m if m != 0.0 else 0.0
    s = v.last_window_frame_std
    return str(int(r * 10) / 10.0) + "fps / " + str(int(m * 1000)) + "ms / " + str(int(s * 1000)) + "ms"


Trace = collections.namedtuple("Trace", ["start_time", "last_time_ref", "min_elapsed", "discard", "all"])


def trace_calls(trace: Trace, frame: types.FrameType, event: str, arg: typing.Any) -> None:
    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    caller = frame.f_back
    assert caller is not None
    caller_line_no = caller.f_lineno
    caller_filename = caller.f_code.co_filename
    t = time.time()
    last_elapsed = t - trace.last_time_ref[0]
    if not trace.discard or trace.discard not in func_filename:
        if last_elapsed > trace.min_elapsed:
            s = "         " if last_elapsed > 0.001 else ""
            trace.all.append(":%s%12.5f %12.5f Call to %s on line %s of %s from line %s of %s" % (
                s, last_elapsed, t - trace.start_time, func_name, func_line_no, func_filename, caller_line_no, caller_filename))
        trace.last_time_ref[0] = t


def begin_trace(min_elapsed: float, discard: typing.Optional[str]) -> Trace:
    trace = Trace(start_time=time.time(), last_time_ref=[0], min_elapsed=min_elapsed, discard=discard, all=list())
    sys.settrace(typing.cast(typing.Callable[[types.FrameType, str, typing.Any], None], functools.partial(trace_calls, trace)))
    return trace


def end_trace(trace: Trace) -> None:
    sys.settrace(None)
    logging.debug("\n".join(trace.all))
    logging.debug("TOTAL: %s", len(trace.all))


@contextlib.contextmanager
def trace(min_elapsed: float = 0.0, discard: typing.Optional[str] = None) -> typing.Generator[None, None, None]:
    t = begin_trace(min_elapsed, discard)
    yield
    end_trace(t)


def sample_stack_all(count: int = 10, interval: float = 0.1) -> None:
    """Sample the stack in a thread and print it at regular intervals."""

    def print_stack_all(l: threading.RLock, ll: typing.MutableSequence[str]) -> None:
        l1 = list()
        l1.append("*** STACKTRACE - START ***")
        code = []
        for threadId, stack in sys._current_frames().items():
            sub_code = []
            sub_code.append("# ThreadID: %s" % threadId)
            for filename, lineno, name, line in traceback.extract_stack(stack):
                sub_code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
                if line:
                    sub_code.append("  %s" % (line.strip()))
            if not "in select" in sub_code[-2] and \
               not "in wait" in sub_code[-2] and \
               not "in _worker" in sub_code[-2] and \
               not "in print_stack_all" in sub_code[-2] and \
               not "in sample_stack_all" in sub_code[-2] and \
               not "in checkcache" in sub_code[-2] and \
               not "do_sleep" in sub_code[-2] and \
               not "sleep" in sub_code[-1] and \
               not any(["in do_sample" in s for s in sub_code]):
                code.extend(sub_code)
        for line in code:
            l1.append(line)
        l1.append("*** STACKTRACE - END ***")
        with l:
            ll.extend(l1)

    def do_sample() -> None:
        l = threading.RLock()
        ll: typing.List[str] = list()
        for i in range(count):
            print_stack_all(l, ll)
            time.sleep(interval)
        with l:
            print("\n".join(ll))

    threading.Thread(target=do_sample).start()


class TraceCloseable:
    all: typing.List[TraceCloseable] = list()

    def __init__(self) -> None:
        self.__tb = traceback.extract_stack()
        TraceCloseable.all.append(self)

    def close(self) -> None:
        TraceCloseable.all.remove(self)

    @classmethod
    def print_leftovers(cls) -> None:
        import sys
        for x in TraceCloseable.all:
            print("**************************")
            print(f"LEAKED {x}")
            print(f"ALLOCATED HERE")
            for line in traceback.StackSummary.from_list(typing.cast(typing.List[typing.Tuple[str, int, str, typing.Optional[str]]], x.__tb)).format():
                print(line, file=sys.stderr, end="")
            print("^^^^^^^^^^^^^^^^^^^^^^^^^^")


class TestEventLoop:
    def __init__(self, event_loop: typing.Optional[asyncio.AbstractEventLoop] = None):
        logging.disable(logging.CRITICAL)  # suppress new_event_loop debug message
        self.__event_loop = event_loop if event_loop else asyncio.new_event_loop()
        logging.disable(logging.NOTSET)

    def close(self) -> None:
        # give cancelled tasks a chance to finish
        self.__event_loop.stop()
        self.__event_loop.run_forever()
        self.__event_loop.run_until_complete(asyncio.gather(*asyncio.all_tasks(loop=self.__event_loop)))
        # now close
        # due to a bug in Python libraries, the default executor needs to be shutdown explicitly before the event loop
        # see http://bugs.python.org/issue28464
        default_executor = getattr(self.__event_loop, "_default_executor", None)
        if default_executor:
            default_executor.shutdown()
        self.__event_loop.close()
        self.__event_loop = typing.cast(asyncio.AbstractEventLoop, None)

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        return self.__event_loop


class Timer:
    def __init__(self, *, threshold: float = 0.0) -> None:
        self.start_time_ns = time.perf_counter_ns()
        self.last_time_ns = self.start_time_ns
        self.threshold = threshold

    def mark(self, title: str) -> None:
        current_time = time.perf_counter_ns()
        if self.threshold == 0.0 or (current_time - self.last_time_ns) // 1E9 > self.threshold:
            print(f"{title}: {(current_time - self.start_time_ns) // 1E6}ms +{(current_time - self.last_time_ns) // 1E6}ms")
        self.last_time_ns = current_time



# reference https://developer.android.com/training/notify-user/build-notification

from __future__ import annotations

# standard libraries
# None

# third party libraries
# None

# local libraries
import typing

from nion.utils import Event
from nion.utils import Observable
from nion.utils import Registry


class Notification(Observable.Observable):
    # click action
    # dismiss
    # auto close

    def __init__(self, notification_type_id: str, task_name: str, title: str, text: str) -> None:
        super().__init__()
        self.dismiss_event = Event.Event()
        self.notification_type_id = notification_type_id
        self.__task_name = task_name
        self.__title = title
        self.__text = text

    @property
    def task_name(self) -> str:
        return self.__task_name

    @task_name.setter
    def task_name(self, value: str) -> None:
        self.__task_name = value
        self.notify_property_changed("task_name")

    @property
    def title(self) -> str:
        return self.__title

    @title.setter
    def title(self, value: str) -> None:
        self.__title = value
        self.notify_property_changed("title")

    @property
    def text(self) -> str:
        return self.__text

    @text.setter
    def text(self, value: str) -> None:
        self.__text = value
        self.notify_property_changed("text")

    def dismiss(self) -> None:
        self.dismiss_event.fire()


class NotificationSourceLike(typing.Protocol):
    notify_event: Event.Event
    notifications: typing.Sequence[Notification]


class NotificationSource(NotificationSourceLike):
    def __init__(self) -> None:
        self.notify_event = Event.Event()
        self.notifications: typing.List[Notification] = list()

    def notify(self, notification: Notification) -> None:
        self.notifications.append(notification)
        self.notify_event.fire(notification)


_notification_source = NotificationSource()


def notify(notification: Notification) -> None:
    _notification_source.notify(notification)


Registry.register_component(_notification_source, {"notification-source"})
